<?php

namespace Drupal\tome_aws;

use Aws\S3\S3Client;
use Drupal;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Extension\ModuleHandler;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\State\StateInterface;
use Drupal\tome_base\PathTrait;
use Drupal\tome_static\StaticGeneratorInterface;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\ClientException;

/**
 * A service to wrap batch operations for deploying to AWS.
 */
class TomeAwsDeployBatch {

  use PathTrait;
  use DependencySerializationTrait;

  /**
   * The static generator.
   *
   * @var \Drupal\tome_static\StaticGeneratorInterface
   */
  protected $static;

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * The config.
   *
   * @var \Drupal\Core\Config\Config
   */
  protected $config;

  /**
   * The config.
   *
   * @var \Drupal\Core\Config\Config
   */
  protected $tomeConfig;

  /**
   * The file system.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandler
   */
  protected $moduleHandler;

  /**
   * @var \Drupal\Core\State\StateInterface
   */
  private $state;

  /**
   * TomeAwsDeployBatch constructor.
   *
   * @param \Drupal\tome_static\StaticGeneratorInterface $static
   *   The static generator.
   * @param \GuzzleHttp\Client $http_client
   *   The HTTP client.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system.
   * @param \Drupal\Core\Entity\EntityTypeManager $entity_type_manager
   * @param \Drupal\Core\Extension\ModuleHandler $module_handler
   */
  public function __construct(StaticGeneratorInterface $static, Client $http_client, ConfigFactoryInterface $config_factory, FileSystemInterface $file_system, EntityTypeManager $entity_type_manager, ModuleHandler $module_handler, StateInterface $state) {

    $this->static            = $static;
    $this->httpClient        = $http_client;
    $this->config            = $config_factory->get('tome_aws.settings');
    $this->tomeConfig        = $config_factory->get('tome_aws.settings');
    $this->fileSystem        = $file_system;
    $this->entityTypeManager = $entity_type_manager;
    $this->moduleHandler     = $module_handler;
    $this->state             = $state;
  }

  /**
   * Determines if the site is configured to properly deploy.
   *
   * @return bool
   *   Whether or not the site is configured to properly deploy.
   */
  public function checkConfiguration() {

    return !empty($this->config->get('aws_s3_client_id')) && !empty($this->config->get('aws_s3_client_secret')) && !empty($this->config->get('aws_s3_bucket_name')) && !empty($this->config->get('aws_s3_bucket_prefix'));
  }

  /**
   * Determines if a static build exists.
   *
   * @return bool
   *   Whether or not a static build exists.
   */
  public function checkStaticBuild() {

    return file_exists($this->static->getStaticDirectory());
  }

  public function getGeneratedBaseURL() {

    return $this->state->get(StaticGeneratorInterface::STATE_KEY_URL, 'http://127.0.0.1');
  }

  /**
   * Gets a batch object for deploying to AWS.
   *
   * @return \Drupal\Core\Batch\BatchBuilder
   *   The batch builder.
   */
  public function getBatch() {

    $batch_builder = new BatchBuilder();

    // 1. Cache remote content metadata
    $this->cacheRemoteContentMetadata();

    // 2. Incremental content Deploy
    $files = $this->getFilesToUpload();
    foreach ($files as $file) {
      $batch_builder->addOperation([$this, 'uploadFile'], [$file]);
    }

    // 3. Set remote redirects
    foreach ($this->getLocalRedirects() as $source => $destination) {
      $batch_builder->addOperation([$this, 'setRemoteRedirect'], [$source, $destination]);
    }

    // todo: 4. Delete obsolete files.

    return $batch_builder;

  }

  public function cacheRemoteContentMetadata() {
    $results = $this->getS3Client()->getPaginator('ListObjects', [
      'Bucket' => $this->config->get('aws_s3_bucket_name')
    ]);
    Drupal::cache('tome_aws_remote_metadata')->invalidateAll();
    foreach ($results as $result) {
      foreach ($result['Contents'] as $object) {
        if ($object['Size'] > 0) {
          $key  = $object['Key'];
          $etag = rtrim(ltrim($object['ETag'], '"'), '"');
          Drupal::cache('tome_aws_remote_metadata')->set($key, $etag);
        }
      }
    }
  }

  public function getFilesToUpload() {
    $files = [];
    foreach ($this->fileSystem->scanDirectory($this->static->getStaticDirectory(), '/.*/') as $file) {
      $relativeFilePath = ltrim($this->getS3BucketPrefix() . str_replace($this->static->getStaticDirectory(), '', $file->uri), '/');
      $remoteETagCache  = Drupal::cache('tome_aws_remote_metadata')->get($relativeFilePath);
      $remoteETag       = $remoteETagCache ? $remoteETagCache->data : NULL;
      $localETag        = md5_file($file->uri);
      if ($remoteETag !== $localETag) {
        $files[] = $file->uri;
      }
    }
    return $files;
  }

  public function getLocalRedirects() {

    if (!$this->moduleHandler->moduleExists('redirect')) {
      return [];
    }

    $redirects_id = $this->entityTypeManager->getStorage('redirect')
                                            ->getQuery()
                                            ->execute()
    ;

    /** @var \Drupal\redirect\Entity\Redirect $redirects_load */
    $redirects_load = $this->entityTypeManager->getStorage('redirect')
                                              ->loadMultiple($redirects_id)
    ;

    $redirects = [];
    foreach ($redirects_load as $redirect_load) {
      $source      = $redirect_load->getSourceUrl();
      $destination = $redirect_load->getRedirectUrl()->toString();

      $redirects[$source] = $destination;
    }

    return $redirects;

  }

  // Remote actions

  /**
   * @param \Drupal\tome_aws\string $local_file_path
   */
  public function uploadFile(string $local_file_path) {

    $file_name = str_replace($this->static->getStaticDirectory(), '', $local_file_path);

    // Must begin without slash
    $remote_file_path = ltrim($this->getS3BucketPrefix() . $file_name, '/');

    // dsm([$local_file_path, $file_name, $remote_file_path]);

    $this->getS3Client()->putObject([
      'Bucket' => $this->config->get('aws_s3_bucket_name'),
      'Key'    => $remote_file_path,
      'Body'   => fopen($local_file_path, 'r'),
      'ACL'    => 'public-read',
    ]);

  }

  /**
   * @param \Drupal\tome_aws\string $source
   * @param \Drupal\tome_aws\string $destination
   */
  public function setRemoteRedirect(string $source, string $destination) {

    // Must begin without slash
    $prepared_source = ltrim($source, '/') . '/index.html';

    $this->getS3Client()->putObject([
      'Bucket'                  => $this->config->get('aws_s3_bucket_name'),
      'Key'                     => $this->getS3BucketPrefix() . $prepared_source,
      'ACL'                     => 'public-read',
      'WebsiteRedirectLocation' => $destination,
    ]);

  }

  //

  private function getS3Client() {
    return new S3Client([
      'region'      => 'eu-central-1',
      'version'     => '2006-03-01',
      'credentials' => [
        'key'    => $this->config->get('aws_s3_client_id'),
        'secret' => $this->config->get('aws_s3_client_secret'),
      ],
    ]);
  }

  private function getS3BucketPrefix($with_trailing_slash = FALSE) {
    return $with_trailing_slash ? $this->config->get('aws_s3_bucket_prefix') : rtrim($this->config->get('aws_s3_bucket_prefix'), '/');
  }

}
